# Copyright (C) 2021, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import pytest
import asyncio
from pathlib import Path
import shutil
from carthage.pytest import *
from carthage.dependency_injection import *
from carthage import base_injector, Machine, setup_task, ConfigLayout, LocalMachine
from carthage.modeling.base import *
from carthage.modeling.implementation import ModelingContainer
from carthage.modeling.decorators import *
from carthage.network import NetworkConfig, Network
from carthage.machine import MachineCustomization, BaseCustomization
import carthage.machine


@pytest.fixture()
def injector():
    injector = base_injector(Injector)
    injector.claim()
    config = injector.get_instance(ConfigLayout)
    base_dir = Path(__file__).parent
    base_dir /= "state"
    config.base_dir = str(base_dir)
    yield injector
    injector.close()
    try:
        shutil.rmtree(base_dir)
    except OSError: pass

@pytest.fixture()
def ainjector(injector):
    return injector(AsyncInjector)


def test_modeling_class_injection(injector):
    class Layout(InjectableModel):
        class nc(NetworkConfig):
            pass
    model = injector(Layout)
    nc = model.injector.get_instance(NetworkConfig)
    assert nc is Layout.nc


def test_namespace_cascade(injector):
    class Layout(InjectableModel):
        green = "green"

        class inner(InjectableModel):
            assert green == "green"
            green = "blue"

        class inner2 (InjectableModel):
            inner = inner

            class MoreInner:
                # This should break the chain because it is not a InjectableModel
                class MostInner(InjectableModel):
                    assert 'inner' not in locals().keys()

    res = injector(Layout.inner)
    assert res.green == "blue"
    assert Layout.inner2.inner is Layout.inner


def test_container(injector):
    class Foo(Injectable):
        pass

    def extra_container(domain):
        d = domain

        @provides(InjectionKey("included", constraint=20))
        class Baz(ModelGroup):
            domain = d

            class server(MachineModel):
                name = "server"
        return Baz

    class Layout(InjectableModel, metaclass=ModelingContainer):

        add_provider(InjectionKey("key1"),
                     42)
        access_key_1 = injector_access("key1")

        net_config = injector_access(InjectionKey(NetworkConfig, domain="evil.com"))

        class RedEnclave(Enclave):

            domain = "evil.com"

            @propagate_up()
            class nc(NetworkConfig):
                pass

            include_container(extra_container)

            @provides("site-network", InjectionKey(Network, name="red"))
            class SiteNetwork(NetworkModel):
                add_provider(InjectionKey(Foo), Foo, propagate=True)
                name = "red"

    res = injector(Layout)
    nc = res.injector.get_instance(InjectionKey(
        NetworkConfig, domain="evil.com", _ready=False))
    assert nc is Layout.RedEnclave.nc
    machine = res.injector.get_instance(InjectionKey(MachineModel, host="server.evil.com", _ready=False))
    assert isinstance(machine, MachineModel)
    assert res.net_config is res.RedEnclave.nc
    assert res.access_key_1 == 42
    foo = res.injector.get_instance(InjectionKey(
        Foo, domain="evil.com", name="red",
        _ready=False))
    assert isinstance(foo, Foo)


def test_dynamic_name(injector):
    class Layout(ModelGroup):

        for i in range(3):
            @dynamic_name(f'i{i+1}')
            class ignored(InjectableModel):
                square = i * i
    assert Layout.i3.square == 4


@async_test
async def test_machine_model(injector):
    class Layout(ModelGroup):

        class Red(Enclave):

            domain = "evil.com"

            class NetConfig (NetworkConfigModel):
                pass

            class Router(MachineModel):
                pass

            class task_machine(MachineModel):

                @setup_task("Frob the frobbables")
                def frob_stuff(self): pass

        red_router = injector_access(InjectionKey(MachineModel, host="router.evil.com"))

    ainjector = injector(AsyncInjector)
    res = await ainjector(Layout)
    assert res.red_router.name == "router.evil.com"
    assert hasattr(res.Red.task_machine, 'setup_tasks')
    await res.Red.task_machine.async_become_ready()


@pytest.mark.xfail(reason="ResolvableModel makes more things async")
@async_test
async def test_example_model(ainjector):
    from carthage.modeling.example import Layout
    res = await ainjector(Layout)
    nc = res.net_config
    assert isinstance(nc, NetworkConfigModel)
    samba = res.injector.get_instance(InjectionKey(MachineModel, host="samba.evil.com"))
    assert samba.network_config is nc


@async_test
async def test_transclusion(ainjector):
    injector = ainjector.injector
    class Moo(carthage.machine.ResolvableModel): pass
    moo = Moo()
    injector.add_provider(InjectionKey(MachineModel, host="moo.com"),
                          moo)
    injector.add_provider(InjectionKey(Machine, host="mar.com"),
                          "baz")

    class FirstLayout(ModelGroup):

        @model_mixin_for(host="mar.com")
        class MarMixin(MachineModel):
            mixed_in = True

    first_layout = await ainjector(FirstLayout)
    injector = first_layout.injector

    class Layout(ModelGroup):

        @transclude_overrides()
        class moo(MachineModel):
            name = "moo.com"

        class mar(*injector(model_bases, "mar.com", MachineModel)):
            name = "mar.com"

    ainjector = injector(AsyncInjector)
    l = await ainjector(Layout)
    assert l.moo is moo
    assert l.mar.machine == "baz"
    assert l.mar.mixed_in == True


@async_test
async def test_generate_and_network(ainjector):

    class Layout(ModelGroup):

        class net(NetworkModel):

            name = "the-net"

        class nc(NetworkConfigModel):
            other_machine = injector_access("other-machine")
            net = injector_access('net')
            add('eth0', net=net, mac=None,
                other=other_machine,
                other_interface="eth 2/1")

        @provides("other-machine")
        class OtherMachine(MachineModel):

            name = "switch.foo.com"

            class fooNetConfig(NetworkConfigModel):
                pass

        class TheMachine(MachineModel):

            name = "machine.foo.com"

            @setup_task("Generate me")
            def generate_stuff(self):
                nonlocal stuff_generated
                stuff_generated = True

            @generate_stuff.invalidator()
            def generate_stuff(self, **kwargs):
                return False

    stuff_generated = False
    l = await ainjector(Layout)
    await l.generate()
    assert stuff_generated
    assert l.TheMachine.network_links['eth0'].other.machine == l.OtherMachine
    assert len(l.net.network_links) == 2


@async_test
async def test_implicit_customization(ainjector):
    class Layout(ModelGroup):

        class foo(MachineModel):
            name = "foo.com"

            class cust(MachineCustomization):
                pass

    l = await ainjector(Layout)
    ainjector.add_provider(machine_implementation_key, dependency_quote(Machine))
    assert hasattr(l.foo.machine_type, 'cust_task')


def test_model_mixin(injector):
    class layout(CarthageLayout):

        @model_mixin_for(host="foo.com")
        class foomixin:
            bar = 42

        class foo(MachineModel):
            name = "foo.com"

    assert layout.foo.bar == 42


@async_test
async def test_local_networking(ainjector):
    class layout(ModelGroup):
        class net1(NetworkModel):
            pass

        class local(MachineModel):
            add_provider(machine_implementation_key, dependency_quote(LocalMachine))

            class nc(NetworkConfigModel):
                net1 = injector_access('net1')

                add('eth0', mac=None,
                    net=net1)
                add('br_net1', net=net1,
                    member='eth0',
                    local_type='bridge',
                    mac=None,
                    )

    l = await ainjector(layout)
    await l.generate()
    assert getattr(l.net1, 'bridge_name', None) == "br_net1"


@async_test
async def test_model_tasks(ainjector):
    class Layout(ModelGroup):
        class tasks(ModelTasks):

            @setup_task("frob stuff")
            def frob_stuff(self):
                nonlocal frobbed
                frobbed = True

            @frob_stuff.check_completed()
            def frob_stuff(self):
                return False

    frobbed = False
    l = await ainjector(Layout)
    assert frobbed is False
    await l.generate()
    assert frobbed is True


@async_test
async def test_tasks_inherit(ainjector):
    class template(MachineModel, template=True):

        class cust1(BaseCustomization):
            @setup_task("Task a")
            def task_1(self):
                nonlocal task_1_run
                task_1_run = True

            @task_1.check_completed()
            def task_1(self):
                return False

    class layout(ModelGroup):
        add_provider(machine_implementation_key, dependency_quote(LocalMachine))

        class local(template, MachineModel):

            class cust2(BaseCustomization):
                @setup_task("task 2")
                def task_2(self):
                    nonlocal task_2_run
                    task_2_run = True

                @task_2.check_completed()
                def task_2(self):
                    return False

    task_1_run = False
    task_2_run = False
    l = await ainjector(layout)
    l.local.machine
    assert task_2_run is False
    await l.local.machine.async_become_ready()
    assert task_2_run
    assert task_1_run
    assert l.local.cust2.task_2


@async_test
async def test_injector_xref_no_cycle(ainjector):
    class layout(ModelGroup):
        class local(MachineModel):
            add_provider(machine_implementation_key, dependency_quote(LocalMachine))

            class cust(BaseCustomization):

                async def async_ready(self):
                    await wait_future
                    return await super().async_ready()

    l = await ainjector(layout)
    wait_future = ainjector.loop.create_future()
    machine_key = InjectionKey("machine")
    l.injector.add_provider(
        machine_key,
        injector_xref(
            InjectionKey(
                MachineModel,
                host="local"),
            InjectionKey(Machine)))
    instantiation_future = ainjector.loop.create_task(l.ainjector.get_instance_async(machine_key))
    await asyncio.sleep(0.2)
    assert instantiation_future.done() is False
    l.local.machine
    wait_future.set_result(True)
    await instantiation_future


@async_test
async def test_globally_unique_with_name(ainjector):
    "Confirm that if @globally_unique shadows inject_by_name the result is globally unique"
    class outer(ModelGroup):
        @provides(InjectionKey("inner", role="role"))
        class inner(ModelGroup):
            @globally_unique_key("comcast_net")
            class comcast_net(NetworkModel):
                pass
    l = await ainjector(outer)
    ainjector = l.ainjector
    await ainjector.get_instance_async('comcast_net')
@async_test
async def test_task_ordering(ainjector):
    "Test that role tasks follow the order of python inheritance"
    class Layout(CarthageLayout):
        class first(MachineModel, template=True):
            class first_cust(BaseCustomization):
                @setup_task("Should run first")
                def first_task(self):
                    nonlocal first_run
                    first_run = True

        class second(MachineModel, template=True):

            class cust(BaseCustomization):
                @setup_task("Should run second")
                def second_task(self):
                    nonlocal second_run
                    assert first_run
                    second_run = True

        # Assume first and second are roles.  The left most (least
        # inherited) role should have its tasks run last
        class machine(second, first):
            pass

    first_run = False
    second_run = False
    ainjector.add_provider(Layout)
    ainjector.add_provider(machine_implementation_key, dependency_quote(LocalMachine))
    layout = await ainjector.get_instance_async(Layout)
    await layout.machine.machine.async_become_ready()
    assert first_run
    assert second_run

@async_test
async def test_async_injector_access(ainjector):
    class layout(CarthageLayout):
        @provides("async")
        class async_injectable(AsyncInjectable):

            async def async_ready(self):
                nonlocal async_ready
                async_ready = True
                await super().async_ready()



    async_ready = False
    l = await ainjector(layout)
    access = injector_access("async")
    await l.ainjector(access)
    assert async_ready is True

@async_test
async def test_detects_class_multi_instantiate(ainjector):
    with pytest.raises(SyntaxError):
        class layout(CarthageLayout):

            @provides("model")
            class a(InjectableModel):
                pass
            add_provider(InjectionKey("b"), a) #should raise

@async_test
async def test_model_subclass_propagation(ainjector):
    "Test that if a template includes classes to be propagated, subclasses of the template properly propagate those classes."
    class TemplateNetwork(NetworkModel):
        @propagate_up()
        class net_config(NetworkConfigModel): pass

    class layout(CarthageLayout):
        @provides(InjectionKey(Network, role='public'))
        class public_network(TemplateNetwork):
            name = 'public_network'
    ainjector.add_provider(layout)
    l = await ainjector.get_instance_async(layout)
    nc = await l.ainjector.get_instance_async(InjectionKey(NetworkConfig, role='public'))
    



# Copyright (C) 2018, 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from pathlib import Path
import contextlib
import os
import os.path
import pkg_resources
import re
import shutil
import sys
import tempfile
import time
from subprocess import check_call, check_output, call
from tempfile import TemporaryDirectory
from .dependency_injection import *
from .config import ConfigLayout, config_key
from . import sh
from .utils import possibly_async, memoproperty
from .setup_tasks import setup_task, SkipSetupTask, SetupTaskMixin, TaskWrapper
import carthage
from .machine import ContainerCustomization, FilesystemCustomization, customization_task


@inject_autokwargs(config_layout=ConfigLayout)
class ContainerVolumeImplementation(AsyncInjectable, SetupTaskMixin):

    def __init__(self, name, path, clone_from=None, **kwargs):
        super().__init__(**kwargs)
        self._path = path
        self._name = name
        self.clone_from = clone_from
        self.closed = False

    @property
    def name(self): return self._name

    @property
    def path(self):
        if self.closed:
            raise RuntimeError("This volume is closed")
        return self._path

    def __del__(self):
        self.close()


@inject_autokwargs(config_layout=ConfigLayout)
class BtrfsVolume(ContainerVolumeImplementation):

    def __repr__(self):
        return "<BtrfsVolume path={}{}>".format(str(self._path),
                                                " closed" if self.closed else "")

    def close(self, canceled_futures=None):
        if self.closed:
            return
        if self.config_layout.delete_volumes:
            subvols = [str(self._path)]
            for root, dirs, files in os.walk(str(self._path), topdown=True):
                device = os.stat(root).st_dev
                for d in dirs:
                    dir = os.path.join(root, d)
                    if os.lstat(dir).st_dev != device:
                        subvols.append(dir)

            for vol in reversed(subvols):
                sh.btrfs('subvolume', 'delete', vol, _bg=False)
        self.closed = True
        super().close(canceled_futures)

    async def async_ready(self):
        try:
            if os.path.exists(self.path):
                try:
                    sh.btrfs("subvolume", "show", str(self.path), _bg=False)
                except sh.ErrorReturnCode:
                    raise RuntimeError("{} is not a btrfs subvolume but already exists".format(self.path))
                # If we're here it is a btrfs subvolume
                return await super().async_ready()
            # directory does not exist
            os.makedirs(os.path.dirname(self.path), exist_ok=True)
            if not self.clone_from:
                await sh.btrfs('subvolume', 'create',
                               str(self.path),
                               _bg=True, _bg_exc=False)
            else:
                await sh.btrfs('subvolume', 'snapshot', str(self.clone_from.path), str(self.path),
                               _bg=True, _bg_exc=False)
            return await super().async_ready()
        except BaseException:
            self.close()
            raise


class ReflinkVolume(ContainerVolumeImplementation):

    async def async_ready(self):
        if self.path.exists():
            return await super().async_ready()
        os.makedirs(self.path.parent, exist_ok=True)
        if self.clone_from:
            await sh.cp(
                "-a", "--reflink=auto",
                str(self.clone_from.path),
                self.path,
                _bg=True, _bg_exc=False)
        else:
            os.mkdir(self.path)
        return await super().async_ready()

    def close(self, canceled_futures=None):
        if self.closed:
            return
        try:
            if self.config_layout.delete_volumes:
                shutil.rmtree(self.path)
        except BaseException:
            pass
        self.closed = True
        super().close(canceled_futures)


@inject(config_layout=ConfigLayout)
class ContainerVolume(AsyncInjectable, SetupTaskMixin):

    def __init__(self, name, *,
                 clone_from=None,
                 implementation=None,
                 config_layout,
                 **kwargs):
        super().__init__(**kwargs)
        path = Path(config_layout.image_dir).joinpath(name)
        os.makedirs(path.parent, exist_ok=True)
        if implementation is None:
            try:
                sh.btrfs(
                    "filesystem", "df", str(path.parent),
                    _bg=False)
                implementation = BtrfsVolume
            except sh.ErrorReturnCode:
                implementation = ReflinkVolume
        self.impl = implementation(name=name,
                                   path=path,
                                   injector=self.injector,
                                   config_layout=config_layout,
                                   clone_from=clone_from)

    async def async_ready(self):
        await self.impl.async_ready()
        await self.populate_volume()
        await super().async_ready()

    async def populate_volume(self):
        "Populate the container volume; called for volumes that exist or are empty"
        return await self.run_setup_tasks()

    @property
    def path(self):
        return self.impl.path

    @property
    def name(self):
        return self.impl.name

    @property
    def config_layout(self): return self.impl.config_layout

    @config_layout.setter
    def config_layout(self, cfg):
        self.impl.config_layout = cfg
        return cfg

    @property
    def stamp_path(self): return self.impl.path

    def __repr__(self):
        return f"<Container {self.impl.__class__.__name__} path:{self.impl.path}>"

    def close(self, canceled_futures=None):
        return self.impl.close(canceled_futures)


class ContainerImage(ContainerVolume):

    async def apply_customization(self, cust_class, method='apply', **kwargs):
        from .container import container_image, container_volume, Container
        injector = self.injector(Injector)
        ainjector = injector(AsyncInjector)
        try:
            injector.add_provider(container_image, dependency_quote(self), close=False)
            injector.add_provider(container_volume, dependency_quote(self), close=False)
            container = await ainjector(Container, name=os.path.basename(self.name), skip_ssh_keygen=True, network_config=None)
            customization = await ainjector(cust_class, apply_to=container, **kwargs)
            if hasattr(customization, 'container_args'):
                container.container_args = customization.container_args
            meth = getattr(customization, method)
            return await meth()
        finally:
            try:
                if container.running:
                    await container.stop_machine()
            except Exception:
                pass
        container.close()

    @setup_task('unpack')
    async def unpack_container_image(self):
        await sh.tar('--xattrs-include=*.*', '-xpzf',
                     self.config_layout.base_container_image,
                     _cwd=self.path,
                     _bg_exc=False,
                     _bg=True)

    @setup_task('cleanup-image')
    def cleanup_image(self):
        try:
            os.unlink(os.path.join(self.path, 'usr/sbin/policy-rc.d'))
        except FileNotFoundError:
            pass
        try:
            os.rename(os.path.join(self.path, "sbin/init.dist"),
                      os.path.join(self.path, "sbin/init"))
        except FileNotFoundError:
            pass


def wrap_container_customization(task: TaskWrapper, **kwargs):
    '''
Takes a :func:`setup_task` and wraps it in a :class:`ContainerCustomization` so that it can be used in a :class:`ContainerVolume`.  Consider the following::

        @setup_task("frob the filesystem")
        async def frob_filesystem(self):
            async with self.filesystem_access() as path: # ...

    Such a setup task cannot be directly applied to a :class:`ContainerVolume` because ContainerVolume does not have *filesystem_access* or any of the other customization methods.  *wrap_container_customization* wraps such a setup_task in a :class:`ContainerCustomization` so that it can be applied to a volume.
    '''
    class cust(ContainerCustomization):
        description = task.description
        task_to_run = task
    return carthage.machine.customization_task(cust, **kwargs)


@inject_autokwargs(config_layout=ConfigLayout)
class ImageVolume(AsyncInjectable, SetupTaskMixin):

    def __init__(self, name, path=None, create_size=None,
                 unpack=None,
                 remove_stamps=False,
                 **kwargs):
        super().__init__(**kwargs)
        name = str(name)  # in case it's a Path
        if path is None:
            path = self.config_layout.vm_image_dir
        self.name = name
        if name.startswith('/'):
            self.path = name
        else:
            self.path = os.path.join(path, name + '.raw')

        if not os.path.exists(self.path):
            os.makedirs(os.path.dirname(self.path), exist_ok=True)
            remove_stamps = True
            if create_size:
                with open(self.path, "w") as f:
                    try:
                        sh.chattr("+C", self.path, _bg=False)
                    except sh.ErrorReturnCode_1:
                        pass
            self.create_size = create_size
        if remove_stamps:
            shutil.rmtree(self.stamp_path, ignore_errors=True)
        os.makedirs(self.stamp_path, exist_ok=True)
        self._pass_self_to_unpack = False
        if callable(unpack):
            self._pass_self_to_unpack = True
            self.unpack = unpack

    def __repr__(self):
        return f"<{self.__class__.__name__} path={self.path}>"

    async def async_ready(self):
        await self.run_setup_tasks()
        return await super().async_ready()

    def delete_volume(self):
        try:
            os.unlink(self.path)
            shutil.rmtree(self.stamp_path)
        except FileNotFoundError:
            pass

    @property
    def stamp_path(self):
        return Path(str(self.path) + '.stamps')

    def close(self, canceled_futures=None):
        if self.config_layout.delete_volumes:
            self.delete_volume()
        super().close(canceled_futures)

    def __del__(self):
        self.close()

    async def unpack(self):
        """Override this method in a subclass to implement alternate creation methods
            Alternatively specify unpack=func() in the constructor
        """
        if self.create_size is None:
            raise RuntimeError(f"{self.path} not found and creation disabled")
        await sh.gzip('-d', '-c',
                      self.config_layout.base_vm_image,
                      _out=self.path,
                      _no_pipe=True,
                      _no_out=True,
                      _out_bufsize=1024 * 1024,
                      _tty_out=False,
                      _bg=True,
                      _bg_exc=False)
        if hasattr(self, 'create_size'):
            os.truncate(self.path, self.create_size)

    @setup_task('unpack')
    async def unpack_installed_system(self):
        if os.path.exists(self.path):
            return  # mark as succeeded rather than skipped

        # We never want to unpack later after we've decided not to
        if self._pass_self_to_unpack:
            return await self.unpack(self)
        return await self.unpack()

    def qemu_config(self, disk_config):
        return dict(
            path=self.path,
            source_type="file",
            driver='raw',
            qemu_source='file',
        )

    async def apply_customization(self, cust_class, method='apply', **kwargs):
        from .container import container_image, container_volume, Container
        injector = self.injector(Injector)
        ainjector = injector(AsyncInjector)
        try:
            image_mount = await ainjector(ContainerImageMount, self)
            injector.add_provider(container_image, image_mount)
            injector.add_provider(container_volume, image_mount)
            container = await ainjector(Container, name=os.path.basename(self.name), skip_ssh_keygen=True, network_config=None)
            customization = await ainjector(cust_class, apply_to=container, **kwargs)
            if hasattr(customization, 'container_args'):
                container.container_args = customization.container_args
            meth = getattr(customization, method)
            return await meth()
        finally:
            try:
                if container.running:
                    await container.stop_machine()
            except Exception:
                pass
            image_mount.close()

    @contextlib.contextmanager
    def image_mounted(self):
        with image_mounted(self.path, mount=False) as i:
            with TemporaryDirectory() as d:
                sh.mount('-osubvol=@',
                         i.rootdev, d, _bg=False)
                i.rootdir = d
                try:
                    yield i
                finally:
                    for i in range(5):
                        try:
                            sh.sync(_bg=False)
                            sh.umount(d,_bg=False)
                            sh.sync(_bg=False)
                            time.sleep(5)
                            break
                        except sh.ErrorReturnCode as e:
                            if 'busy' in e.stderr.lower():
                                time.sleep(0.5)
                            else:
                                raise

    def clone_for_vm(self, name, *,
                     path=None, volume_type='qcow'):
        kwargs = {}
        if path is not None:
            kwargs['path'] = path
        if volume_type == 'qcow':
            return self.injector(QcowCloneVolume, name, self, **kwargs)
        elif volume_type == 'raw':
            return self.injector(ImageVolume, name=name, image_base=self.path, **kwargs)
        else:
            raise ValueError("Unknown volume type {}".format(volume_type))


class BlockVolume(ImageVolume):

    def __init__(self, path, **kwargs):
        if 'name' in kwargs:
            raise ValueError('BlockVolume does not take name even though ImageVolume does')
        assert path.startswith('/')
        assert path[1] != '/'  # otherwise stamp_path needs to be more clever
        super().__init__(name=path, unpack=False)

    @memoproperty
    def stamp_path(self):
        res = Path(self.config_layout.state_dir) / "block_volume_stamps" / self.path[1:]
        os.makedirs(res, exist_ok=True)
        return res

    def qemu_config(self, disk_config):
        res = dict(
            source_type='block',
            path=self.path,
            driver='raw',
            qemu_source='dev')
        return res


@inject_autokwargs(
    config_layout=ConfigLayout,
)
class QcowCloneVolume(Injectable):

    def __init__(self, name, volume, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.path = os.path.join(self.config_layout.vm_image_dir, name + ".qcow")
        if not os.path.exists(self.path):
            sh.qemu_img(
                'create', '-fqcow2',
                '-obacking_file=' + volume.path,
                '-obacking_fmt='+volume.qemu_config(dict())['driver'],
                str(self.path),_bg=False)

    def delete_volume(self):
        try:
            os.unlink(self.path)
        except FileNotFoundError:
            pass

    def close(self, canceled_futures=None):
        if self.config_layout.delete_volumes:
            self.delete_volume()

    def qemu_config(self, disk_config):
        return dict(
            path=self.path,
            source_type='file',
            driver='qcow2',
            qemu_source='file'
        )


@inject_autokwargs(config_layout=ConfigLayout)
class ContainerImageMount(AsyncInjectable, SetupTaskMixin):

    '''Mount a disk image for use as a container_image or
    container_volume.  Note that this works for LVM and raw images, but
    not currently for qcow2 images.  If you use this as a
    container_volume, you must call close yourself as the container will
    not.
    '''

    def __init__(self, image,
                 **kwargs):
        super().__init__(**kwargs)
        self.image = image

        self.name = image.name
        self.mount_image()
        self.stamp_path = image.stamp_path

        def __repr__(self):
            return f"<{self.__class__.__name__} for {repr(self.image)}>"

    def mount_image(self):
        self.mount_context = self.image.image_mounted()
        self.mount = self.mount_context.__enter__()

    def unmount_image(self):
        try:
            mount_context = self.mount_context
            del self.mount_context
            del self.mount
            mount_context.__exit__(None, None, None)
        except AttributeError:
            pass

    def close(self):
        self.unmount_image()
        try:
            del self.image
        except AttributeError:
            pass

    async def async_ready(self):
        try:
            await self.run_setup_tasks()
            return await super().async_ready()
        except BaseException:
            self.close()
            raise

    def __del__(self):
        self.close()

    @property
    def path(self):
        return self.mount.rootdir


@inject(
    ainjector=AsyncInjector,
    config_layout=ConfigLayout)
def image_factory(name, image_type='raw',
                  image=ImageVolume, *,
                  config_layout, ainjector):
    assert image_type == 'raw'
    path = os.path.join(config_layout.vm_image_dir, name + '.raw')
    return ainjector(image, name=name, path=path,
                     create_size=config_layout.vm_image_size)


class SshAuthorizedKeyCustomizations(FilesystemCustomization):

    description = "Set up authorized_keys file"

    @setup_task('Copy in hadron-operations ssh authorized_keys')
    @inject(authorized_keys=carthage.ssh.AuthorizedKeysFile)
    def add_authorized_keys(self, authorized_keys):
        os.makedirs(os.path.join(self.path, "root/.ssh"), exist_ok=True)
        shutil.copy2(authorized_keys.path,
                     os.path.join(self.path, 'root/.ssh/authorized_keys'))


class image_mounted(object):

    '''A contextmanager for mounting and image'''

    def __init__(self, image=None, mount=True, extra_dirs=[],
                 subvol="@", rootdir=None):
        self.image = image
        self.mount = mount
        self.mount_dirs = "dev proc sys dev/pts".split()
        self.mount_dirs.extend(extra_dirs)
        self.rootdir = rootdir
        if rootdir and image:
            raise RuntimeError("Do not specify both image and rootdir")
        self.loopdev = None
        self.subvol = subvol

    def __enter__(self):
        if self.image:
            self.clear_existing_loops(self.image)
            if self.image.startswith('/dev'):
                # This technically won't work for an lvm image unless the
                # calling user calls kpartx themselves, but it will work
                # for an actual disk.  Detecting the LVM case is kind of
                # tricky.
                if os.path.exists(self.image + '2'):
                    self.rootdev = self.image + '2'
                else:
                    self.rootdev = self.image
            else:
                mappings = str(check_output(['kpartx',
                                             '-asv', self.image]), 'utf8').split("\n")
                partitions = list(map(lambda x: x.split()[2], filter(lambda x: x, mappings)))
                self.loopdev = re.match(r'^(loop[0-9]+)', partitions[0]).group(1)
                self.loopdev = "/dev/" + self.loopdev
                partitions = list(map(lambda x: "/dev/mapper/" + x, partitions))
                self.rootdev = partitions[1]
        try:
            self.rootdir = tempfile.mkdtemp()
            if self.mount:
                options = []
                if self.subvol:
                    options.append('-osubvol=' + self.subvol)
                check_call(['mount'] + options + [self.rootdev, self.rootdir])
                for dir in self.mount_dirs:
                    check_call(['mount', '-obind', '/' + dir,
                                os.path.join(self.rootdir, dir)])
        except Exception as e:
            self.__exit__(type(e), e, None)
            raise e
        return self

    def clear_existing_loops(self, image):
        lines = str(check_output(['losetup', '-j', image]), 'utf-8').split("\n")
        if not lines:
            return
        for line in lines:
            (loopdev, *rest) = line.split(':')
            if not loopdev.startswith('/dev'):
                continue
            call(['kpartx', '-ds', loopdev])
            call(['losetup', '-d', loopdev])
        call(['kpartx', '-d', self.image])

    def __exit__(self, *excinfo):
        if self.rootdir and self.mount:
            for dir in reversed(self.mount_dirs):
                call(['umount',
                      os.path.join(self.rootdir, dir)])
            call(['umount', self.rootdir])
            os.rmdir(self.rootdir)
        if self.loopdev:
            check_call(['kpartx', '-d', self.loopdev])
            call(['losetup', '-d', self.loopdev])
        return False

    def chroot(self, command, *args):
        if isinstance(command, str):
            command = [command]
        command.extend(args)
        command = ['chroot', self.rootdir] + command
        check_call(command)


__all__ = ('ContainerVolume', 'ContainerImage',
           "wrap_container_customization",
           'SetupTaskMixin',
           'SkipSetupTask',
           'ImageVolume',
           'SshAuthorizedKeyCustomizations')



# Copyright (C) 2018, 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import logging
import os
import os.path
from pathlib import Path
from carthage.image import ImageVolume, SetupTaskMixin, setup_task
from carthage.utils import memoproperty
from carthage.dependency_injection import *
from carthage.config import ConfigSchema, ConfigLayout, config_key, ConfigAccessor
from ..config.types import ConfigPath
from .. import sh
from .credentials import vmware_credentials
from .datastore import VmwareDataStore
from .datacenter import VmwareDatacenter


logger = logging.getLogger('carthage.vmware')

#: Injection key used for looking up templates for VMs
image_datastore_key = InjectionKey(VmwareDataStore, role="images")
vm_datastore_key = InjectionKey(VmwareDataStore, role='vm')


@inject_autokwargs(
    config_layout=ConfigLayout,
    store=image_datastore_key
)
class VmdkTemplate(SetupTaskMixin, AsyncInjectable):

    '''
    Produce a VMDK from an image that can be loaded as a template VM.


    :param: image
        A :class:`~carthage.image.ImageVolume` to turn into a VMDK template.


    :param dspath: The base path within the datastore at which images should be stored; if ``None``, taken from ``vmware.image_datastore.path`` in config.


'''

    def __init__(self, image, dspath=None,
                 prefix="",
                 **kwargs):
        self.image = image
        self.dspath = dspath
        super().__init__(**kwargs)
        if dspath is None:
            self.dspath = self.config_layout.vmware.image_datastore.path
            if self.dspath and not self.dspath.endswith('/'):
                self.dspath += '/'
        assert str(image.path).endswith('.raw')
        path = Path(image.path)
        self.prefix = prefix
        if not self.prefix.endswith('/'):
            self.prefix += '/'
        self.paths = [path.stem + ".vmdk", path.stem + "-flat.vmdk"]
        self.path = path.parent
        self.stamp_path = self.image.stamp_path

    def __repr__(self):
        return f"<VMDK for \"{self.image.path}\" datastore={self.store.name}>"

    @setup_task("generate-vmdk")
    async def generate_vmdk(self):
        await self.image.async_become_ready()
        await sh.qemu_img(
            'convert',
            "-Ovmdk",
            "-osubformat=monolithicFlat",
            self.image.path,
            str(self.path / self.paths[0]),
            _bg=True, _bg_exc=False)

        return self

    @generate_vmdk.check_completed()
    def generate_vmdk(self):
        try:
            st = os.stat(self.path / self.paths[0])
            return st.st_mtime
        except FileNotFoundError:
            return False

    @setup_task("copy-vmdk")
    async def copy_vmdk(self):
        store = self.store
        for p in self.paths:
            await store.copy_in(self.path / p, self.dspath + self.prefix + p)

    @memoproperty
    def disk_path(self):
        return f'[{self.store.name}]{self.dspath}{self.prefix}{self.paths[0]}'


class VmwareDatastoreConfig(ConfigSchema, prefix="vmware.datastore"):
    name: str
    #: Path within the data store at which VMs are stored
    path: str = ""
    local_path: ConfigPath = None


class ImageDatastoreConfig(ConfigSchema, prefix="vmware.image_datastore"):
    name: str
    #: Base path within the data store at which images are stored
    path: str = "templates"
    local_path: ConfigPath = None


@inject(injector=Injector)
async def produce_datastore_from_config(prefix, *, injector):
    ainjector = injector(AsyncInjector)
    config = injector(ConfigAccessor, prefix=prefix)
    return await ainjector(VmwareDataStore, name=config.name, local_path=config.local_path)

__all__ = ('image_datastore_key', 'vm_datastore_key',
           'VmdkTemplate',
           'VmwareDataStore')



# Copyright (C) 2019, 2021, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import os.path
from pathlib import Path
from tempfile import TemporaryDirectory
from .dependency_injection import *
from . import ConfigLayout, sh
from .ssh import RsyncPath, SshKey, rsync
from .setup_tasks import *
__all__ = []
rsync_supports_mkpath_state = None


def rsync_supports_mkpath():
    global rsync_supports_mkpath_state
    if rsync_supports_mkpath_state is None:
        rsync_supports_mkpath_state = 'mkpath' in sh.rsync('--help')
    return rsync_supports_mkpath_state


@inject(config=ConfigLayout,
        ainjector=AsyncInjector)
async def rsync_git_tree(git_tree, target: RsyncPath,
                         *, config, ainjector):
    '''
    Copy a git tree into a target system.

Clone the ``HEAD`` of a Git working copy into a new temporary directory  This preserves committed files but does not preserve untracked or uncommitted files.  Rsync that directory to the path on a remote system indicated by *target*.
'''

    assert isinstance(target, RsyncPath)
    rsync_opts = []
    if rsync_supports_mkpath():
        rsync_opts.append('--mkpath')
    if rsync_supports_mkpath() and not str(target.path).endswith('/'):
        target = RsyncPath(target.machine, str(target.path) + '/')
    git_tree = sh.git('rev-parse', '--show-toplevel', _cwd=git_tree)
    git_tree = str(git_tree.stdout, 'utf-8').rstrip()
    dir = None
    try:
        dir = TemporaryDirectory(dir=config.state_dir)
        await sh.git('clone',
                     git_tree, dir.name,
                     _bg=True, _bg_exc=False)
        return await ainjector(rsync, '-a', '--delete',
                               *rsync_opts,
                               dir.name + '/', target)
    finally:
        dir.cleanup()

__all__ += ['rsync_git_tree', ]


def git_tree_hash(git_tree):
    '''
    Return the HEAD of a git tree suitable for use in a setup_task's hash function
'''
    res = sh.git('rev-parse', 'HEAD',
                 _cwd=git_tree,
                 )
    return str(res.stdout, 'utf-8').rstrip()


__all__ += ['git_tree_hash']


def git_checkout_task(url, repo):
    '''Returns a :func:`setup_task` that will checkout a give git repository.
The resulting setup_task has an attribute *repo_path* which is a function returning the path to the repo

'''
    @inject(config=ConfigLayout)
    def repo_path(config):
        checkouts = Path(config.checkout_dir)
        return checkouts / repo

    @setup_task(f"Checkout {repo} repository")
    @inject(injector=Injector)
    async def checkout_repo(self, injector):
        if callable(url):
            url_resolved = injector(url)
        else:
            url_resolved = url
        return await checkout_git_repo(url_resolved, repo, injector=injector)

    @checkout_repo.invalidator()
    @inject(injector=Injector)
    def checkout_repo(self, last_run, injector):
        path = injector(repo_path)
        return path.exists()
    checkout_repo.repo_path = repo_path
    return checkout_repo


@inject(injector=Injector)
def checkout_git_repo(url, repo, *, injector):
    config = injector(ConfigLayout)
    path = Path(config.checkout_dir) / repo
    os.makedirs(config.checkout_dir, exist_ok=True)
    if path.exists():
        return sh.git("pull", "--rebase",
                      _bg=True,
                      _bg_exc=False,
                      _cwd=path)
    else:
        return sh.git("clone",
                      url, str(path),
                      _bg=True,
                      _bg_exc=False)


__all__ += ['git_checkout_task', 'checkout_git_repo']



# Copyright (C) 2019, 2020, 2021, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import os
import os.path
import pytest
import machine_mock
from carthage.dependency_injection import *
import carthage
import carthage.ansible
from carthage.pytest import *

from carthage.modeling import *

state_dir = os.path.join(os.path.dirname(__file__), "test_state")


@pytest.fixture()
@async_test
async def configured_ainjector(ainjector):
    config = ainjector.injector(carthage.ConfigLayout)
    config.state_dir = state_dir
    ainjector.add_provider(carthage.ssh.SshKey)
    enable_modeling_ansible(ainjector)
    return ainjector


@async_test
async def test_run_play(configured_ainjector):
    ainjector = configured_ainjector
    res = await ainjector(carthage.ansible.run_play, [carthage.ansible.localhost_machine],
                          [{"debug": "msg=foo"}])


@async_test
async def test_run_failing_play(configured_ainjector):
    ainjector = configured_ainjector
    with pytest.raises(carthage.ansible.AnsibleFailure):
        res = await ainjector(carthage.ansible.run_play, [carthage.ansible.localhost_machine],
                              [{"fail": "msg=foo"}])


@async_test
async def test_ansible_with_log(configured_ainjector):
    ainjector = configured_ainjector
    try:
        os.unlink(state_dir + "/ansible.log")
    except BaseException:
        pass
    await ainjector(carthage.ansible.run_play,
                    [carthage.ansible.localhost_machine],
                    {'debug': 'msg=barbaz'},
                    log=state_dir + "/ansible.log")
    with open(state_dir + "/ansible.log", "rb") as f:
        log_contents = f.read()
    assert b'barbaz' in log_contents


@async_test
async def test_inventory(configured_ainjector):
    ainjector = configured_ainjector

    class Layout(ModelGroup):

        add_provider(machine_implementation_key, dependency_quote(machine_mock.Machine))
        domain = "example.com"

        class m1(MachineModel):

            ansible_vars = dict(foo=90)
            ansible_groups = ['bar', 'baz']

    layout = await ainjector(Layout)
    ainjector = layout.injector.get_instance(AsyncInjector)
    inventory = await ainjector(carthage.ansible.AnsibleInventory, os.path.join(state_dir, "inventory.yml"))
    for g in layout.m1.ansible_groups:
        assert 'm1.example.com' in inventory.inventory[g]['hosts']
        assert inventory.inventory['all']['hosts']['m1.example.com']['foo'] == 90



# Copyright (C) 2018, 2019, 2020, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage.pytest import *
import pytest
import posix
from ipaddress import *
from carthage import base_injector
import carthage.network.config
from carthage.network import Network, BridgeNetwork, V4Config
from carthage.dependency_injection import *
from carthage.modeling import *


@pytest.fixture()
def injector():
    if posix.geteuid() != 0:
        pytest.skip("Not running as root; network tests skipped", )
    return base_injector.claim()


@async_test
async def test_network_create(injector, loop):
    ainjector = injector(AsyncInjector)
    net = await ainjector(Network, name="testnet")
    net.close()


@pytest.mark.xfail(reason="Needs adjusting for namespace")
@async_test
async def test_network_veth(injector, loop):
    ainjector = injector(AsyncInjector)
    net = await ainjector(Network, name="testnet")
    net = await net.access_by(BridgeNetwork)
    ve = net.add_veth('database.hadronindustries.com')
    net.close()


@async_test
async def test_v4_config_secondary_expand(ainjector):
    '''Test v4_config with deferred elements and secondary addresses
    '''
    def address():
        return '10.1.0.1'
    class layout(CarthageLayout):
        @provides("net")
        class net(NetworkModel):
            name = 'net_1'
            v4_config = V4Config(network='10.0.0.0/8')

        class machine(MachineModel):
            class net_config(NetworkConfigModel):
                add('eth0', net=net,
                    mac=None,
                    v4_config=V4Config(
                        secondary_addresses = [address],
                        ))

    l = await ainjector(layout)
    config = l.machine.network_links['eth0'].merged_v4_config
    assert config.secondary_addresses == [carthage.network.config.SecondaryAddress(private=IPv4Address(address()))]
    assert IPv4Address(address()) in config.network

    
        



# Copyright (C)  2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.
import typing

from carthage.modeling import *
import carthage.modeling.implementation
from .base import *
from carthage import *
from carthage.dependency_injection import *
from carthage.oci import OciImage

__all__ = []


class PodmanPodModel( carthage.machine.NetworkedModel, InjectableModel, metaclass=carthage.modeling.implementation.ModelingContainer):

    '''A container that can group a number of :class:`MachineModels` representing Podman containers.

    * By default, ``machine_implementation_key`` within a :class:`PodmanPodModel` is :class:`PodmanContainer`.

    * By default, when added to an enclosing injector, this model does not provide :class:`PodmanPod` in that context, although it does provide :class:`PodmanPod` within its own injector.  An example illustrates::

        class Layout(CarthageLayout):

            #Even though this layout includes  PodmanPods, PodmanPod is not provided in the injector of the layout.  That means containers outside of a PodmanPodModel will not be associated with a pod by default

            class pod1(PodmanPodModel):

                # These containers are within pod1 and because within its own scope pod1 provides PodmanPod, the containers will be added to the pod.

                class container1(MachineModel): pass

                class container2(MachineModel): pass

            # But this container is not in a pod.

            class container3(MachineModel): pass

    In contrast to using PodmanPodModel, :class:`PodmanPod` can be used along side containers in another modeling grouping.  When used this way :class:`PodmanPod` does provide :class:`PodmanPod` in the enclosing injector::

        class layout(CarthageLayout):

            class group1(ModelGroup):

                # Everything in this group is pod1

                class pod1(PodmanPod): name = 'pod1'

                class container1(MachineModel): pass

    '''

    add_provider(machine_implementation_key, dependency_quote(PodmanContainer))
    self_provider(carthage.machine.network_namespace_key)

    @classmethod
    def name_for(cls):
        return getattr(cls, 'name', cls.__name__)
    
    def __init__(self, **kwargs):
        self.name = self.name_for()
        super().__init__(**kwargs)
        self.network_links = {}
        pod_key = InjectionKey(PodmanPod, name=self.name, _globally_unique=self.pod_name_global)
        if pod_key in self.ignored_by_transclusion:
            self.injector.add_provider(InjectionKey(PodmanPod), injector_access(pod_key), close=False)
        else:
            class Pod(PodmanPod):
                name = self.name
            self.injector.add_provider(InjectionKey(PodmanPod), Pod)

    def __init_subclass__(cls, template=False, **kwargs):
        super().__init_subclass__(**kwargs)
        if not template:
            cls.add_provider(InjectionKey(PodmanPod, name=cls.name_for(), _globally_unique=cls.pod_name_global),
                         injector_access(InjectionKey(PodmanPod)),
                         close=False,
                         propagate=cls.pod_name_global,
                         transclusion_overrides=cls.pod_name_global)
            globally_unique_key(InjectionKey(carthage.machine.ResolvableModel, name=cls.name_for()+'-pod'))(cls)
                                         
                                          
    pod_name_global = True  # : If True, the pod name is globally unique

    @classmethod
    def our_key(self):
        name = self.name_for()
        return InjectionKey(self.__class__, name=name)

    @classmethod
    def supplementary_injection_keys(self, k):
        if k.constraints:
            yield InjectionKey(PodmanPod, **k.constraints)
            yield InjectionKey(PodmanPodModel, **k.constraints)
            if 'name' in k.constraints and not issubclass(k.target, carthage.machine.ResolvableModel):
                yield InjectionKey(carthage.machine.ResolvableModel, name=k.constraints['name']+'-pod',
                                   _globally_unique=self.pod_name_global)

    pod = injector_access(InjectionKey(PodmanPod))

    def __repr__(self):
        return f'<{self.__class__.__name__} name:{self.name}>'
    

__all__ += ['PodmanPodModel']


class PodmanImageModel(ImageRole, PodmanImage):

    '''
    Like a :class:`PodmanImage` excetp:

    * This is an :class:`InjectableModel` so modeling language constructs can be used

    * Any :class:`FilesystemCustomization` or :class:`ContainerCustomization` that are registered with the injector are automatically treated as image layers **after** any explicit setup_tasks.

    '''

    # We need to have some of the attributes from AbstractMachineModel so that start_machine can work
    override_dependencies: typing.Union[bool, Injector, Injectable, InjectionKey] = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.injected_tasks_added = False
        self.injector.add_provider(InjectionKey(PodmanImageModel), self)
        self.network_links = {}

    def add_injected_tasks(self):
        for k, customization in self.injector.filter_instantiate(
                carthage.machine.BaseCustomization, [
                    'description'], stop_at=self.injector):
            if issubclass(customization, (
                    carthage.machine.ContainerCustomization,
                    carthage.machine.FilesystemCustomization,
                    )):
                self.add_setup_task(image_layer_task(customization))
            else:
                logger.warn(f'{customization} is an inappropriate customization for {self}')
        self.injected_tasks_added = True

    async def build_image(self):
        if not self.injected_tasks_added:
            self.add_injected_tasks()
        return await super().build_image()

    def __init_subclass__(cls, **kwargs):
        if issubclass(cls, MachineModel):
            raise TypeError(cls.__name__+' should not be both a PodmanImageModel and MachineModel.  This probably means you tried to add a role that is not an ImageRole to a PodmanImageModel')
        super().__init_subclass__(**kwargs)
        
        

__all__ += ['PodmanImageModel']

class ContainerfileImageModel(ContainerfileImage, InjectableModel):

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.oci_image_tag:
            provides(InjectionKey(OciImage, tag=cls.oci_image_tag))(cls)
                     

__all__ += ['ContainerfileImageModel']




# Copyright (C)  2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import asyncio
import dataclasses
import enum
from pathlib import Path
import random
import traceback
import uuid
import yaml

from entanglement.interface import *
from entanglement.memory import *
from entanglement.filter import *
from entanglement.types import register_type, register_enum


from ..dependency_injection import InjectionKey, Injector, is_obj_ready, get_dependencies_for
from ..setup_tasks import SetupTaskMixin, _iso_time

__all__ = []

def encode_injection_key(k):
    # Note that encoding and decoding an InjectionKey is not an identity.  In particular, the type will become a string.
    # This may or may not be reasonable depending on how we end up using this code.
    target = k.target
    if isinstance(target, type):
        target = target.__name__
    return [target, k.constraints]

def decode_injection_key(v):
    return InjectionKey(v[0],**v[1])
register_type(InjectionKey, encode_injection_key, decode_injection_key)

class CarthageDestination(FilteredSyncDestination):

    filter_should_listen_returns_true = True
    
    def __init__(self):
        super().__init__(dest_hash=random.randbytes(32),
                         name='websocket to ip address should go here')
        self.add_filter(Filter(lambda o: True, registry=carthage_registry))
        
class CarthageRegistry(SyncStoreRegistry):

    def associate_with_manager(self, manager):
        self.manager = manager # This should be weak according to docs
        super().associate_with_manager(manager)
        
    def injector_id(self, injector):
        return id(injector)
    
    def on_add_provider(self, scope,  other_keys, target_key, inspector,
                        **kwargs):
        injector_id = self.injector_id(scope)
        injector_info = self.get_or_create(InjectorInfo, injector_id)
        if scope.parent_injector:
            injector_info.parent_id = id(scope.parent_injector)
        try: injector_info.description = repr(scope)
        except Exception: pass
        self.store_synchronize(injector_info)
        for k in other_keys | {target_key}:
            o = ProvidedDependency(injector_id=injector_id, key=k, provider_id=inspector.provider_id)
            self.store_synchronize(o)
        provider_info = self.get_or_create(ProviderInfo, inspector.provider_id)
        try: provider_info.value_repr = repr(inspector.get_value_no_instantiate())
        except Exception:
            provider_info.value_repr = None
        self.store_synchronize(provider_info)
        for dependency in get_dependencies_for(inspector.get_value_no_instantiate(), inspector.injector):
            sd = StaticDependency(
                provider_id=inspector.provider_id,
                key=dependency.key,
                provided_by=dependency.provider_id)
            self.store_synchronize(sd)

    def on_update(self, event, target, **kwargs):
        inspector = target
        provider_info = self.get_or_create(ProviderInfo, inspector.provider_id)
        try: provider_info.provider_repr = repr(inspector.get_value_no_instantiate())
        except Exception: pass
        if event == 'dependency_progress':
            provider_info.state = InstantiationProgress.in_progress

        else: #dependency_final
            if is_obj_ready(inspector.get_value(ready=False)):
                provider_info.state = InstantiationProgress.ready
            else: provider_info.state = InstantiationProgress.not_ready
        try: provider_info.value_injector_id = id(inspector.get_value_no_instantiate().injector)
        except Exception: pass
        self.store_synchronize(provider_info)
        value = inspector.get_value_no_instantiate()
        if isinstance(value, SetupTaskMixin):
            asyncio.ensure_future(self.handle_tasks(value))

    async def handle_task(self, inspector, running=False, should_run=None, exception=None):
        task_info = self.get_or_create(TaskInfo, inspector.instance_id, inspector.stamp)
        task_info.running = running
        if should_run is None:
            try: should_run = await inspector.should_run(ainjector=None)
            except Exception: pass
        if should_run is not None: task_info.should_run = should_run
        try:         task_info.last_run = _iso_time(inspector.last_run)
        except AttributeError: task_info.last_run = None
        task_info.description = inspector.description
        if exception:
            task_info.last_failure = str(exception)
        self.store_synchronize(task_info)
        for subinspector in inspector.subtasks():
            await self.handle_task(subinspector)
            

    async def handle_tasks(self, obj):
        for inspector in obj.inspect_setup_tasks():
            await self.handle_task(inspector)

    async def on_task_event(self, event, target, task, **kwargs):
        exception = None
        should_run = None
        if event == 'task_start': running = True
        else: running = False
        if event == 'task_ran': should_run = False
        if event == 'task_fail':
            exception = kwargs['exception']
        for inspector in  target.inspect_setup_tasks():
            if inspector.task == task:
                return await self.handle_task(
                    inspector,
                    should_run=should_run,
                    exception=exception,
                    running=running)

    def on_instantiation_failed(self, target, scope, target_key, **kwargs):
        failed_instantiation = self.get_or_create(FailedInstantiation,
                                                  id(scope), target_key)
        failed_instantiation.exception = "".join(traceback.format_exception(target.exception))
        failed_instantiation.dependency = target.dependency
        self.store_synchronize(failed_instantiation)
        
    def dump(self, path):
        path = Path(path)
        result: dict[str, list] = {}
        for t, store in self.stores_by_class.items():
            objects = []
            for o in store.values():
                sync_repr = o.to_sync()
                sync_repr['_sync_type'] = o.sync_type
                objects.append(sync_repr)
            result[t.sync_type] = objects
        path.write_text(yaml.dump(result, default_flow_style=False ))
        

    def instrument_injector(self, injector):
        injector.add_event_listener(InjectionKey(Injector), {'add_provider'}, self.on_add_provider)
        injector.add_event_listener(
            InjectionKey(Injector), {'dependency_progress', 'dependency_final'}, self.on_update)
        injector.add_event_listener(
            InjectionKey(SetupTaskMixin),
            {'task_ran', 'task_start', 'task_fail', 'task_should_run'},
            self.on_task_event)
        injector.add_event_listener(
            InjectionKey(Injector), {'dependency_instantiation_failed'}, self.on_instantiation_failed)
        for k, inspector in injector.inspect():
            self.on_add_provider(scope=injector, target_key=inspector.key, other_keys=set(),
                                 inspector=inspector
                                 )
            

carthage_registry = CarthageRegistry()
carthage_registry.register_syncable(SyncOwner.sync_type, SyncOwner)

__all__ += ['carthage_registry']

@dataclasses.dataclass
class ProvidedDependency(StoreInSyncStoreMixin):

    key: InjectionKey = sync_property(constructor=True)
    injector_id: int = sync_property(constructor=True)
    provider_id: int = sync_property(None)

    sync_registry = carthage_registry
    sync_primary_keys = ('injector_id', 'key')

__all__ += ['ProvidedDependency']

class InstantiationProgress(enum.Enum):
    not_instantiated = 0
    in_progress = 1
    not_ready = 2
    ready = 3

__all__ += ['InstantiationProgress']
register_enum(InstantiationProgress)


@dataclasses.dataclass
class ProviderInfo(StoreInSyncStoreMixin):

    id: int = sync_property(constructor=True)
    state: InstantiationProgress= sync_property(InstantiationProgress.not_instantiated)
    
    value_repr:str = sync_property(None)
    value_injector_id: int = sync_property(None)
    

    sync_registry = carthage_registry
    sync_primary_keys = ('id',)
    

__all__ += ['ProviderInfo']

@dataclasses.dataclass
class StaticDependency(StoreInSyncStoreMixin):

    #: The ProviderInfo to which this dependency is attached
    provider_id:int = sync_property(constructor=True)
    key: InjectionKey = sync_property(constructor=True)
    #: The ProviderInfo that provides this dependency
    provided_by:int = sync_property(None)

    sync_primary_keys = ('provider_id', 'key')
    sync_registry = carthage_registry

__all__ += ['StaticDependency']

@dataclasses.dataclass
class InjectorInfo(StoreInSyncStoreMixin):

    id: int = sync_property(constructor=True)
    parent_id: int = sync_property(None)
    description: str = sync_property(None)

    sync_primary_keys = ('id',)
    sync_registry = carthage_registry

@dataclasses.dataclass
class TaskInfo(StoreInSyncStoreMixin):

    instance_id: id = sync_property(constructor=True)
    stamp: str = sync_property(constructor=True)
    description: str = sync_property("")
    running: bool = sync_property(False)
    should_run: bool = sync_property(None)
    last_run: str = sync_property(None)
    last_failure: str = sync_property(None)
    

    sync_primary_keys = ('instance_id', 'stamp')
    sync_registry = carthage_registry
    
__all__ += ['TaskInfo']

def encode_fi_dependency(val):
    return [encode_injection_key(x) for x in val]

def decode_fi_dependency(val):
    return [decode_injection_key(x) for x in val]

@dataclasses.dataclass
class FailedInstantiation(StoreInSyncStoreMixin):

    injector_id: int = sync_property(constructor=True)
    key: InjectionKey = sync_property(constructor=True)
    exception: str = sync_property("")
    dependency: list[InjectionKey] = sync_property(
        dataclasses.field(default_factory=lambda: []),
        encoder=encode_fi_dependency,
        decoder=decode_fi_dependency)
    sync_primary_keys = ('injector_id', 'key')
    sync_registry = carthage_registry

__all__ += ['FailedInstantiation']



# Copyright (C)  2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import dataclasses
from carthage.dependency_injection import *
from .setup_tasks import setup_task, SetupTaskMixin
from .utils import memoproperty
from .config.types import ConfigPath
import carthage.machine


__all__ = []

#: This key provides a string whose result is the container image name to pull  or an image ID that is locally available.  Examples include ``docker.io/library/debian:latest`` or ``debian:unstable``
oci_container_image = InjectionKey('oci/container_image')

__all__ += ['oci_container_image']


@inject_autokwargs(
    oci_read_only=InjectionKey('oci_read_only', _optional=NotPresent)
)
class OciManaged(SetupTaskMixin, AsyncInjectable):

    #:Should this object be treated as read only
    oci_read_only = False

    async def find(self):
        '''Returns falsy if the object does not exist.  Ideally returns the creation time in unix time, otherwise returns True if the creation time cannot be determined.
        '''
        raise NotImplementedError

    @setup_task("Construct Object", order=400)
    async def find_or_create(self):
        # Make sure we call find.  When called by setup_tasks,
        # check_completed runs first and so we do not need to call, but
        # for an explicit call we need to call find ourselves.
        if not hasattr(self, '_find_result'):
            self._find_result = await self.find()
        if self._find_result:
            return  # find was successful
        del self._find_result
        if self.oci_read_only:
            raise RuntimeError(f'{self} is read only but does not exist')

        await self.do_create()
        self._find_result = await self.find()
        return self._find_result

    @find_or_create.check_completed()
    async def find_or_create(self):
        self._find_result = await self.find()
        return self._find_result

    async def do_create(self):
        raise NotImplementedError

    def __repr__(self):
        res = f'<{self.__class__.__name__} '
        try: res += f'name:{self.name} '
        except Exception: pass
        try:
            if self.id:
                res += f'id: {self.id} '
        except Exception: pass
        return res+'>'
    

__all__ += ['OciManaged']


@dataclasses.dataclass
class OciExposedPort(Injectable):

    container_port: int
    host_ip: str = "0.0.0.0"
    host_port: int = ""
    proto: str = 'tcp'

    def default_instance_injection_key(self):
        if self.proto != 'tcp':
            return InjectionKey(OciExposedPort, container_port=self.container_port, proto=self.proto)
        return InjectionKey(OciExposedPort, container_port=self.container_port)


__all__ += ['OciExposedPort']


@inject_autokwargs(
    oci_interactive=InjectionKey('oci_interactive', _optional=NotPresent),
    oci_tty=InjectionKey('oci_tty', _optional=NotPresent),
)
class OciContainer(OciManaged):

    #: Should stdin be kept open?
    oci_interactive = False

    #: Allocate a tty for stdio
    oci_tty = False

    @memoproperty
    def exposed_ports(self):
        '''Return a sequence of :class:`OciExposedPort` for any container ports that should be exposed.

        By default, instantiate all *OciExposedPort* instances in the injector.
        '''
        result = self.injector.filter_instantiate(OciExposedPort, ['container_port'])
        return [i[1] for i in result]

    @memoproperty
    def mounts(self):
        '''
        Sequence of :class:`OciMount` objects to be mounted to the container.

        By default instantiate *OciMount* objects in the injector hierarchy.
        '''
        results = self.injector.filter_instantiate(OciMount, ['destination'])
        return [i[1] for i in results]

    @memoproperty
    def oci_command(self):
        '''Override the container command if non-None.
        Defaults to checking on the model
'''
        if hasattr(self.model, 'oci_command'):
            return self.model.oci_command
        return None


__all__ += ['OciContainer']


class OciImage(OciManaged):

    def __init__(self, *, oci_image_tag=None, id=None, **kwargs):
        if oci_image_tag:
            self.oci_image_tag = oci_image_tag
        if id:
            self.id = id
        if not hasattr(self, 'oci_image_tag') and not hasattr(self, 'id'):
            raise TypeError('Either oci_image_tag or id is required')
        super().__init__(**kwargs)

    oci_image_author = ""
    oci_image_command = None
    oci_image_entry_point = None
    id = None


__all__ += ['OciImage']


@dataclasses.dataclass
class OciMount(Injectable):

    destination: str
    source: str
    options: str = ""
    mount_type: str = 'volume'

    def default_instance_injection_key(self):
        return InjectionKey(OciMount, destination=self.destination)

    @classmethod
    def default_class_injection_key(self):
        return InjectionKey(OciMount, destination=self.destination)

    def source_resolved(self, injector):
        return injector(ConfigPath, self.source)


__all__ += ['OciMount']


class OciPod(OciManaged):

    #: The name of the pod
    name: str = None
    #: the ID of the pod
    id: str = None

    def __init__(self, name=None, id=None, **kwargs):
        if name:
            self.name = name
        if id:
            self.id = id
        if not (self.name or self.id):
            raise TypeError('Either name or id is mandatory')
        super().__init__(**kwargs)
        if self.id:
            self.oci_read_only = True

    @memoproperty
    def exposed_ports(self):
        '''Return a sequence of :class:`OciExposedPort` for any container ports that should be exposed.

        By default, instantiate all *OciExposedPort* instances in the injector.
        '''
        result = self.injector.filter_instantiate(OciExposedPort, ['container_port'])
        return [i[1] for i in result]


__all__ += ['OciPod']


@dataclasses.dataclass
class OciEnviron(Injectable):

    assignment: str
    scope: str = 'all'  # : or exec or image or container

    def default_instance_injection_key(self):
        if self.scope == 'all':
            return InjectionKey(OciEnviron, name=self.name)
        else:
            return InjectionKey(OciEnviron, name=self.name, scope=self.scope)

    @property
    def name(self):
        name, sep, value = self.assignment.partition('=')
        return name


__all__ += ['OciEnviron']


def host_mount(dir, readonly=False):
    options = []
    if readonly:
        options.append('ro=true')
    return OciMount(dir, dir, mount_type='bind', options=','.join(options))


__all__ += ['host_mount']

container_host_model_key = InjectionKey(carthage.machine.AbstractMachineModel, role='container_host')

__all__ += ['container_host_model_key']
